@@ -35,6 +35,7 @@ from functools import partial
 
 import numpy as np
 from numpy.testing import assert_, run_module_suite, assert_allclose
+import pytest
 
 # disable the MC progress bar
 import os
@@ -718,11 +719,11 @@ class TestMESolveSuperInit:
         avg_diff = np.mean(abs(actual_answer - expt) / actual_answer)
         assert_(avg_diff < 1e-5)
 
+
 class TestMESolverMisc:
     """
     A test class for the misc mesolve features.
     """
-
     def testMEFinalState(self):
         "mesolve: final_state has correct dims"
 
@@ -757,6 +758,55 @@ class TestMESolverMisc:
                          options=opts)
         assert_(psi0.dims == result.final_state.dims)
 
+    def test_num_collapse_set(self):
+        H = sigmaz()
+        psi = (basis(2, 0) + basis(2, 1)).unit()
+        ts = [0, 1]
+        for c_ops in (
+            sigmax(),
+            [sigmax()],
+            [sigmay(), sigmax()],
+        ):
+            res = qutip.mesolve(H, psi, ts, c_ops=c_ops)
+            if not isinstance(c_ops, list):
+                c_ops = [c_ops]
+            assert res.num_collapse == len(c_ops)
+
+    # All Hamiltonians should have dimensions [3, 2], so "bad" states can have
+    # [2, 3] instead - the aim is to test mismatch in most cases.
+    @pytest.mark.parametrize(['state'], [
+        pytest.param(basis([2, 3], [0, 0]), id='ket_bad_tensor'),
+        pytest.param(basis([2, 3], [0, 0]).proj(), id='dm_bad_tensor'),
+        pytest.param(basis([2, 3], [0, 0]).dag(), id='bra_bad_tensor'),
+        pytest.param(to_super(basis([2, 3], [0, 0]).proj()),
+                     id='super_bad_tensor'),
+        pytest.param(basis([3, 2], [0, 0]).dag(), id='bra_good_tensor'),
+        pytest.param(operator_to_vector(basis([3, 2], [0, 0]).proj()),
+                     id='operket_good_tensor'),
+        pytest.param(operator_to_vector(basis([3, 2], [0, 0]).proj()).dag(),
+                     id='operbra_good_tensor'),
+        pytest.param(tensor(basis(2, 0), qeye(3)), id='nonsquare_operator'),
+    ])
+    @pytest.mark.parametrize(['operator'], [
+        pytest.param(qeye([3, 2]), id='constant_hamiltonian'),
+        pytest.param(liouvillian(qeye([3, 2]), []), id='constant_liouvillian'),
+        pytest.param([[qeye([3, 2]), lambda t, args: 1]], id='py_scalar'),
+        pytest.param(lambda t, args: qeye([3, 2]), id='py_hamiltonian'),
+        pytest.param(lambda t, args: liouvillian(qeye([3, 2]), []),
+                     id='py_liouvillian'),
+    ])
+    def test_incorrect_state_caught(self, state, operator):
+        """
+        Test that mesolve will correctly catch an input state that is not a
+        correctly shaped state, relative to the Hamiltonian or Liouvillian.
+
+        Regression test for gh-1456.
+        """
+        times = [0, 1e-5]
+        c_op = qeye([3, 2])
+        with pytest.raises(ValueError):
+            mesolve(operator, state, times, c_ops=[c_op])
+
 
 class TestMESolveStepFuncCoeff:
     """
@@ -870,5 +920,55 @@ class TestMESolveStepFuncCoeff:
             msg="evolution with feedback not proceding as expected")
 
 
+def test_non_hermitian_dm():
+    """Test that mesolve works correctly for density matrices that are
+    not Hermitian.
+    See Issue #1460
+    """
+    N = 2
+    a = destroy(N)
+    x = (a + a.dag())/np.sqrt(2)
+    H = a.dag() * a
+
+    # Create non-Hermitian initial state.
+    rho0 = x*fock_dm(N, 0)
+
+    tlist = np.linspace(0, 0.1, 2)
+
+    options = Options()
+    options.store_final_state = True
+    options.store_states = True
+
+    result = mesolve(H, rho0, tlist, e_ops=[x], options=options)
+
+    msg = ('Mesolve is not working properly with a non Hermitian density' +
+       ' matrix as input. Check computation of '
+      )
+
+    imag_part = np.abs(np.imag(result.expect[0][-1]))
+    # Since we used an initial state that is not Hermitian, the expectation of
+    # x must be imaginary for t>0.
+    assert_(imag_part > 0,
+            msg + "expectation values. They should be imaginary")
+
+    # Check that the output state is not hermitian since the input was not
+    # Hermitian either.
+    assert_(not result.final_state.isherm,
+            msg + " final density  matrix. It should not be hermitian")
+    assert_(not result.states[-1].isherm,
+            msg + " states. They should not be hermitian.")
+
+    # Check that when suing a callable we get imaginary expectation values.
+    def callable_x(t, rho):
+        "Dummy callable_x expectation operator."
+        return expect(rho, x)
+    result = mesolve(H, rho0, tlist, e_ops=callable_x)
+
+    imag_part = np.abs(np.imag(result.expect[-1]))
+    assert_(imag_part > 0,
+            msg + "expectation values when using callable operator." +
+            "They should be imaginary.")
+
+
 if __name__ == "__main__":
     run_module_suite()
